/***************************************************************************
	file:			emuI2C.c
	product:	API-002 V1.1
	descr:		Diese API emuliert eine I2C (TWI) Schnittstelle und ermoeglicht 
						das Senden und Empfangen der Daten ueber die I2C (TWI) Schnittstelle 
						eines 8051 AT89C5131A Mikrocontrollers in einem Singel Master - Multi Slave System.
	date: 		10.2013 - first release

	version 	1.0	- first release
						1.1	- masterNack() hinzugefuegt
								- masterNack() in EmuI2C_ReceiveData() implementiert
								- EmuI2C_StartConditon() Ueberpruefung ob Bus frei ist

	compiler:	Keil Cx51 V9.05

	mc:				AT89C5131A
	crystal:	12MHz, 24MHz

	author: 	Viktor Schabelski info@approbe.de
	licence:	Creative Commons Lizenz
                http://www.stack2Learn.de/cc.html

**************************************************/

/*-----------------------------------------------------------------------------
	includes
---------------------------------------------------*/
#include "emuI2C.h"
#include "emuI2C_config.h"


/*-----------------------------------------------------------------------------
 *	global/constant variables																									*
 *----------------------------------------------------------------------------*/
#ifndef LOW
#define LOW 0
#endif

#ifndef HIGH
#define HIGH 1
#endif

#define _SDA_HIGH 	(sda |= HIGH)
#define _SDA_LOW 		(sda &= LOW)
#define _SCL_HIGH 	(scl |= HIGH)
#define _SCL_LOW 		(scl &= LOW)

// Takt
static const unsigned char DELAY = 2;


/*-----------------------------------------------------------------------------
	prototype private functions
---------------------------------------------------*/
static void EmuI2C_Delay(unsigned char );
static void EmuI2C_Cycle(void);

static void EmuI2C_BusFree (void);
static void EmuI2C_StartCondition(void);
static void EmuI2C_StopCondition(void);
static void EmuI2C_MasterAck(void);
static void EmuI2C_MasterNack(void);

static errorCodeEmuI2C_t EmuI2C_SendByte(unsigned char );
static unsigned char EmuI2C_ReceiveByte(void);


/*-----------------------------------------------------------------------------
	api
---------------------------------------------------*/
/*=====================================================================*//**
*	@brief		delay
************************************************************************/
static void EmuI2C_Delay(unsigned char time){
	while(time--);
}


static void EmuI2C_Cycle(void){
	_SCL_HIGH;
	EmuI2C_Delay(DELAY);
	_SCL_LOW;
}

/*=====================================================================*//**
*	@brief		ruheLage von I2C
************************************************************************/
static void EmuI2C_BusFree (void){
		_SCL_HIGH;
		_SDA_HIGH;
}

/*=====================================================================*//**
*	@brief		start condition
************************************************************************/
static void EmuI2C_StartConditon(void){
	if((sda == LOW)||(scl == LOW)){
		EmuI2C_MasterNack();
		EmuI2C_StopCondition();
		EmuI2C_Delay(5);
		EmuI2C_BusFree();
	}
	_SDA_LOW;
	EmuI2C_Delay(DELAY);
	_SCL_LOW;
}

/*=====================================================================*//**
*	@brief		stop condition
************************************************************************/
static void EmuI2C_StopCondition(void) {
	_SDA_LOW;
	EmuI2C_BusFree();
}

/*=====================================================================*//**
*	@brief		master ackn
************************************************************************/
static void EmuI2C_MasterAck(void) {
	_SDA_LOW;
	EmuI2C_Cycle();
	_SDA_HIGH;
}

/*=====================================================================*//**
*	@brief		master not ackn
************************************************************************/
static void EmuI2C_MasterNack(void) {
	_SDA_HIGH;
	EmuI2C_Cycle();
}

/*=====================================================================*//**
*	@brief		sende ein Byte
*
*	@param		dataByte - Byte zum Senden
*	@return		EMU_I2C_TRANSMISSION_FAIL
*	@return		EMU_I2C_SUCCESS
 ************************************************************************/
static errorCodeEmuI2C_t EmuI2C_SendByte(unsigned char dataByte) {
	unsigned char mask = 0x80, i;
	
	for (i = 0; i < 8; i++) {
		if (mask & dataByte){
			_SDA_HIGH;
		}else{
			_SDA_LOW;
		}
		// Daten freigeben
		EmuI2C_Cycle();
		// naechsten Bit vorbereiten
		mask >>= 1;
	}
	_SDA_HIGH;	
	EmuI2C_Delay(1);
	_SCL_HIGH;
	EmuI2C_Delay(1);
	
	//Fehlererkennung
	if (sda) {
		_SCL_LOW;		
			return (EMU_I2C_TRANSMISSION_FAIL);
	} else {
		_SCL_LOW;
		return (EMU_I2C_SUCCESS);
	}
}


/*=====================================================================*//**
*	@brief		empfange ein Byte
*
* @return		byteRX - empfangener Byte
************************************************************************/
static unsigned char EmuI2C_ReceiveByte(void) {
	unsigned char byteRX = 0x00;
	unsigned char i = 0x00;
	//unsigned char wert = 128;

	_SDA_HIGH;
	_SCL_LOW;

	for(i = 0; i < 8; i++){
		byteRX <<= 1;
		EmuI2C_Delay(DELAY);
		_SCL_HIGH;		
		byteRX |= sda;
		EmuI2C_Delay(DELAY);
		_SCL_LOW;
		
		
//		EmuI2C_Delay(DELAY);
//		_SCL_HIGH;
//		if(sda) byteRX = byteRX+wert;
//		wert = wert/2;
//		
//		EmuI2C_Delay(DELAY);
//		_SCL_LOW;
		
	}
	_SCL_LOW;
	return byteRX;
}


/*=====================================================================*//**
*	@brief		Diese Funktion sendet ein Array ueber die I2C (TWI) Schnittstelle.
*
*		An diese Funktion wird ein Array mit Daten uebergeben.
*		Die Slave Adresse des I2C Bausteins soll in dem Arrayindex [0] stehen. 
*		Die weiteren Daten zum Senden stehen in anderen Indizes des Arrays. 
*		Die Variable "size" teilt der Funktion die Anzahl der Bytes die gesendet 
*		werden muessen mit.
*
*	Bsp. 
*	unsigned char pData[size] = {SLAVE_ADDRESS, daten};
*
*	emuI2C_sendData(pData, size);
*
*	@param		size - Array Groesse
*	@param		pData - Pointer auf Array
*	@return		EMU_I2C_TRANSMISSION_FAIL
*	@return		EMU_I2C_SUCCESS
************************************************************************/
errorCodeEmuI2C_t emuI2C_sendData(unsigned char* pData, unsigned char size){
	errorCodeEmuI2C_t error = EMU_I2C_BUSY;
	unsigned char errorFlag = 3;
	
	EmuI2C_StartConditon();

	while (size){
		error = EmuI2C_SendByte(*pData);
		size--;
		pData++;

// Fehlerbehandlung
		if (error != EMU_I2C_SUCCESS) {
			EmuI2C_StopCondition();
			return (EMU_I2C_TRANSMISSION_FAIL);		
		}else{
			error = EMU_I2C_SUCCESS;
		} 	
	}	
	EmuI2C_StopCondition();
	return (error);
}


/*=====================================================================*//**
*	@brief		Diese Funktion empfaengt Daten ueber die I2C (TWI) Schnittstelle.
*
*		An diese Funktion wird ein Array mit Daten uebergeben. 
*		Die Slave Adresse + 1 (R/W  Bit) des I2C Bausteins soll in dem Arrayindex [0] stehen. 
*		Die Variable "size" teilt der Funktion die Anzahl der Bytes die empfangen werden muessen mit.
*
*		Die empfangenen Bytes werden in den gleichen Array ab Index [0] gespeichert.
*
*
*	Bsp. 
*	unsigned char pData[size] = {SLAVE_ADDRESS+1, daten};
*
*	EmuI2C_ReceiveData(pData, size);
*
*
* @param		size - bytes quantity to be received
*	@param		pData - array with data, uchData[0] = SlaveAddress
************************************************************************/
errorCodeEmuI2C_t EmuI2C_ReceiveData(unsigned char* pData, unsigned char size) {
	errorCodeEmuI2C_t error = EMU_I2C_BUSY;
	
	EmuI2C_StartConditon();
	
	// send slave address + Read(1)
	error = EmuI2C_SendByte(pData[0]);
 	if(error != EMU_I2C_SUCCESS){
 		EmuI2C_StopCondition();
 		return EMU_I2C_TRANSMISSION_FAIL;
 	}else{
 		error = EMU_I2C_SUCCESS;
 	}
	

/* receive data via emuI2C */
	while (size) {
		*pData = EmuI2C_ReceiveByte();

		// Master Ack nach allen Bytes senden
		// Nach dem letzten Byte Master not Ack senden
		if (size > 1){
			EmuI2C_MasterAck();
			pData++;
		}else{
			EmuI2C_MasterNack();
		}
		size--;
	}
	EmuI2C_StopCondition();
	return error;
}